import textwrap
from copy import deepcopy
from itertools import chain

import openai
from loguru import logger
from pydantic import BaseModel

from promuet.match_items import ListMatchItem, TemplateMatchItem, VariableMatchItem, VarType, parse_template
from promuet.prompt_message import PromptMessage


class Prompt(BaseModel):
    messages: list[PromptMessage]
    response_format: str
    is_continuation: bool = False
    description: str = ''

    def __init__(
        self,
        messages: list[PromptMessage],
        response_format: str,
        *,
        is_continuation: bool = False,
        description: str = '',
    ):
        super().__init__(
            messages=messages,
            response_format=textwrap.dedent(response_format),
            is_continuation=is_continuation,
            description=description,
        )

    def execute(
        self, variables: dict[str, VarType], prior_messages: list[PromptMessage] | None = None
    ) -> dict[str, VarType]:
        template = TemplateMatchItem(self.response_format)
        var_names = [x.var_name for x in template.children if isinstance(x, VariableMatchItem | ListMatchItem)]
        if all(x in variables for x in var_names):
            return {x: variables[x] for x in var_names}
        prior_messages = (prior_messages or []) if self.is_continuation else []
        if self.is_continuation and not prior_messages:
            msg = 'Expected continuation of prior messages'
            raise ValueError(msg)
        chat_messages = [x.to_chat_message(variables) for x in chain(prior_messages, self.messages)]
        response = openai.chat.completions.create(model='gpt-3.5-turbo', messages=chat_messages)
        response_text = response.choices[0].message.content
        response_text = response_text.strip()
        if response_text.startswith('```\n') and response_text.endswith('\n```'):
            response_text = response_text.removeprefix('```\n').removesuffix('\n```')
        return parse_template(self.response_format, response_text)


class PromptFlow(BaseModel):
    prompts: list[Prompt]

    def __init__(self, *prompts: Prompt):
        super().__init__(prompts=list(prompts))

    def run(self, variables: dict[str, VarType]) -> dict[str, VarType]:
        prior_messages = []
        variables = deepcopy(variables)
        all_new_vars = {}
        for prompt in self.prompts:
            if not prompt.is_continuation:
                prior_messages.clear()
            logger.info("Running prompt '{}'...", prompt.description)
            new_vars = prompt.execute(variables, prior_messages)
            logger.debug('Extracted variables:\n\n{}', new_vars)
            variables.update(new_vars)
            all_new_vars.update(new_vars)
            prior_messages.extend(prompt.messages)
        return all_new_vars



import json
from pathlib import Path
import sys
from typing import Optional

from openai import OpenAI

from .prompt_flow import PromptFlow, VarType


class AppFlow:
    def __init__(self, client: OpenAI = OpenAI(), session_path: Optional[str] = None, debug: bool = False):
        self.client = client
        self.session_path = Path(session_path) if session_path else None
        self.session: dict[str, VarType] = self.load_session()
        self.debug = debug

    def load_session(self) -> dict[str, VarType]:
        return (
            {}
            if not self.session_path or not self.session_path.is_file()
            else json.loads(self.session_path.read_text())
        )

    def save_session(self):
        if self.session_path:
            self.session_path.write_text(json.dumps(self.session, indent=4))

    def prompt_from_cli(self, field_name: str, prompt_message: str):
        """Gets a field from session or prompts the user for it."""
        if field_name not in self.session:
            self.session[field_name] = input(prompt_message).rstrip('.')

    def run_prompt(self, prompt: PromptFlow, extra_args: dict = None):
        self.session.update(extra_args or {})
        new_vars = prompt.run(self.session)
        self.session.update(new_vars)
        self.save_session()
        for key, value in new_vars.items():
            if key + '.str' in self.session:
                value = self.session[key + '.str']
            if key.endswith('.str'):
                continue

            assert isinstance(value, str), (key, value, key + '.str')
            self._debug('\n### {} ###'.format(key))
            self._debug(value)
            self._debug()
    
    def __getitem__(self, key: str) -> VarType:
        return self.session[key]

    def __setitem__(self, key: str, value: VarType):
        self.session[key] = value
        self.save_session()
    
    def __delitem__(self, key: str):
        del self.session[key]
        self.save_session()
    
    def __contains__(self, key: str) -> bool:
        return key in self.session
    
    def _debug(self, *args, **kwargs):
        if self.debug:
            sys.stdout.write(*args, **kwargs)
            sys.stdout.flush()



import json
import textwrap

from promuet.match_items import TemplateMatchItem, extract_data_vars


def test_match_items():
    template = TemplateMatchItem(
        """
            Task {{task_number:int}}: {{task_title}}
            Description: {{task_description}}

            Verification: {{task_verification_items:list}}
        """
    )
    input_string = """
        Task 12: Clean the house
        Description:
        I need to clean the house to make it nice.

        The floors need to be cleaned, the rugs vacuumed, and the laundry folded.

        Verification:
        - Floors should have no dust
        - Rugs should be free of dirt
        - Laundry should be folded
    """
    data = template.parse(input_string)
    assert data['.str'].strip() == textwrap.dedent(input_string).strip()
    data = extract_data_vars(data)
    assert data == {
        'task_number': 12,
        'task_title': 'Clean the house',
        'task_description': 'I need to clean the house to make it nice.\n\nThe floors need to be cleaned, the rugs vacuumed, and the laundry folded.',
        'task_verification_items': [
            'Floors should have no dust',
            'Rugs should be free of dirt',
            'Laundry should be folded',
        ],
    }


def test_match_item_list():
    template = TemplateMatchItem(
        """
            Here are the tasks relating to '{{relates_to}}':
            [[:tasks:]]
            Task {{num:int}}: {{title}}
            [[:tasks:]]
            Note: {{note:str}}
        """
    )
    input_string = """
        Here are the tasks relating to 'hamburgers':

        Task 1: Do something

        Task 2: Something else

        Note: These are the tasks.
    """
    rdata = template.parse(input_string)
    data = extract_data_vars(rdata)
    assert data == {
        'relates_to': 'hamburgers',
        'tasks': [{'num': 1, 'title': 'Do something'}, {'num': 2, 'title': 'Something else'}],
        'note': 'These are the tasks.',
    }, json.dumps(rdata, indent=4)


def test_list():
    template = TemplateMatchItem(
        """
            Features:
            [[:app_mvp_features:]]
            {{number:int}}. {{title}}:
            [[:details:]]
            - {{content}}
            [[:details:]]
            [[:app_mvp_features:]]

            Note: ...
        """
    )
    input_string = """
Features:
1. Virtual Hacking World:
- Create a basic virtual environment with a single hacking scenario for players to explore.
- Include a graphical interface to depict the virtual system, with interactive elements to manipulate.
- Integrate some basic hacking challenges, such as bypassing firewalls or cracking passwords.

2. Coding Challenge:
- Develop a simple coding challenge within the hacking scenario for players to solve.
- Provide a basic code editor where players can write and execute code to progress through the challenge.
- Incorporate a scoring system to evaluate the player's coding skills and progression.

3. Realistic Hacking Tools:
- Include a limited set of realistic hacking tools that players can utilize during their hacking mission.
- Implement basic functionalities of these tools, such as port scanning or packet sniffing.
- Ensure that the tools have a visually appealing and intuitive user interface.

Note: ...
"""

    data = template.parse(input_string)
    features = data['app_mvp_features']
    assert len(features) == 3
    assert [x['title'] for x in features] == [
        'Virtual Hacking World',
        'Coding Challenge',
        'Realistic Hacking Tools',
    ]
    assert [x['content'] for x in features[0]['details']] == [
        'Create a basic virtual environment with a single hacking scenario for players to explore.',
        'Include a graphical interface to depict the virtual system, with interactive elements to manipulate.',
        'Integrate some basic hacking challenges, such as bypassing firewalls or cracking passwords.',
    ]


if __name__ == '__main__':
    test_match_items()
    test_match_item_list()
    test_list()



import textwrap
from typing import Literal

from pydantic import BaseModel

from promuet.match_items import VarType, serialize_template


class PromptMessage(BaseModel):
    role: Literal['system', 'user', 'assistant']
    content: str

    def to_chat_message(self, variables: dict[str, VarType] | None = None) -> dict:
        return {
            'role': self.role,
            'content': serialize_template(self.content, variables or {}),
        }


class SystemMessage(PromptMessage):
    def __init__(self, content: str):
        super().__init__(role='system', content=textwrap.dedent(content))


class UserMessage(PromptMessage):
    def __init__(self, content: str):
        super().__init__(role='user', content=textwrap.dedent(content))


class AssistantMessage(PromptMessage):
    def __init__(self, content: str):
        super().__init__(role='assistant', content=textwrap.dedent(content))



from __future__ import annotations

import json
import re
import textwrap
from typing import Any

from loguru import logger

VarType = int | list[str] | dict[str, 'VarType'] | str


class RegexBuilder:
    def __init__(self):
        self.counter = 0

    def new_capture_group(self, inner_regex: str) -> tuple[str, str | None]:
        self.counter += 1
        group_name = f'group_{self.counter}'
        return f'(?P<{group_name}>{inner_regex})', group_name


class MatchItem:
    def __init__(self, template: str | None, _regex_builder: RegexBuilder):
        self.template = template
        self.match_regex = ''
        self.group_name: str | None = None

    def extract(self, match: re.Match) -> dict[str, VarType]:
        raise NotImplementedError

    def parse(self, input_string: str) -> dict[str, VarType]:
        regex = r'^\s*' + self.match_regex + r'\s*$'
        m = re.match(regex, input_string, re.DOTALL)
        if not m:
            msg = (
                'Input string does not match template!\n\n'
                f'Template:\n{self.template}\n\n'
                f'Template regex:\n{regex}\n\n'
                f'Input string:\n{input_string}\n'
            )
            raise ValueError(msg)
        return self.extract(m)


class LiteralMatchItem(MatchItem):
    def __init__(self, content: str, regex_builder: RegexBuilder):
        super().__init__(None, regex_builder)
        # TODO: Handle edge cases (ie. \\ at end of line)
        self.match_regex = re.sub(
            r'(?:\\\n)+',
            '\\\n+',
            re.escape(content).replace(r'\ ', r'\s+').replace('\\n', '\\\n'),
        )
        self.group_name = None

    def extract(self, _match: re.Match) -> dict[str, VarType]:
        return {}


class VariableMatchItem(MatchItem):
    template_regex = r'\{\{(?P<var_name>\w+)(?::(?P<type>\w+))?\}\}'
    inner_regexes = {'int': r'\d+', 'list': r'.*?', 'str': r'.*?'}

    def __init__(self, template_match: re.Match, regex_builder: RegexBuilder):
        super().__init__(template_match.string, regex_builder)
        self.var_name = template_match.group('var_name')
        self.var_type = template_match.group('type') or 'str'
        regex = self.inner_regexes[self.var_type]
        self.match_regex, self.group_name = regex_builder.new_capture_group(regex)

    def extract(self, match: re.Match) -> dict[str, VarType]:
        value = match.group(self.group_name)
        parsed_value = self._parse(value)
        new_vars = {self.var_name: parsed_value}
        if isinstance(parsed_value, list):
            new_vars[self.var_name + '.str'] = value
        return new_vars

    def _parse(self, value: str) -> VarType:
        if self.var_type == 'int':
            return int(value)
        if self.var_type == 'list':
            # Handles numbered bullets (e.g., 1. item), dash bullets (e.g., - item),
            # asterisk bullets (e.g., * item), and plus bullets (e.g., + item).
            bullet_pattern = re.compile(r'^\s*(\d+\.|[-\*\+])\s*(.*)$', re.M)
            matches = bullet_pattern.findall(value)

            if matches:
                # Extract the actual content (second group) from the matches
                return [match[1].strip() for match in matches]
            # Fallback to split by newline if no pattern is matched
            return [item.strip() for item in value.split('\n') if item.strip()]
        if self.var_type == 'str':
            return value
        raise AssertionError


class ListMatchItem(MatchItem):
    template_regex = r'\[\[:(?P<list_name>\w+):\]\](?P<template_content>.*?)\[\[:\2:\]\]'

    def __init__(self, template_match: re.Match, regex_builder: RegexBuilder):
        super().__init__(template_match.string, regex_builder)
        self.var_name = template_match.group('list_name')
        self.item_template = TemplateMatchItem(template_match.group('template_content'), regex_builder)
        regex, self.group_name = regex_builder.new_capture_group(f'(?:{self.item_template.match_regex})+')
        self.match_regex = regex

    def remove_named_groups(self, pattern: str) -> str:
        return re.sub(r'\(\?P<[^>]+>', '(?:', pattern)

    def extract(self, match: re.Match) -> dict[str, VarType]:
        items = []
        original_text = match.group(self.group_name)

        pattern = self.item_template.match_regex
        lookahead_pattern = self.remove_named_groups(pattern)
        combined_pattern = f'({pattern})(?=(?:{lookahead_pattern}|$))'

        pos = 0
        while pos < len(original_text):
            item_match = re.search(combined_pattern, original_text[pos:], re.DOTALL)
            if item_match is None:
                break

            items.append(self.item_template.extract(item_match))
            pos += item_match.end(1)

        return {self.var_name: items, self.var_name + '.str': original_text}


class TemplateMatchItem(MatchItem):
    def __init__(self, template: str, regex_builder: RegexBuilder = None):
        regex_builder = regex_builder or RegexBuilder()
        template = textwrap.dedent(template).strip()
        super().__init__(template, regex_builder)
        unified_pattern = '|'.join(
            f'(?P<{cls.__name__}>{cls.template_regex})' for cls in [ListMatchItem, VariableMatchItem]
        )
        self.children: list[MatchItem] = []
        last_pos = 0
        for match in re.finditer(unified_pattern, template, re.DOTALL):
            if match.start() != last_pos:
                literal = LiteralMatchItem(template[last_pos : match.start()], regex_builder)
                self.children.append(literal)

            for cls in [ListMatchItem, VariableMatchItem]:
                if match.group(cls.__name__):
                    child = cls(match, regex_builder)
                    self.children.append(child)
                    break
            last_pos = match.end()
        if last_pos != len(template):
            literal = LiteralMatchItem(template[last_pos:], regex_builder)
            self.children.append(literal)

        combined_regex = ''.join([child.match_regex for child in self.children])
        self.match_regex, self.group_name = regex_builder.new_capture_group(combined_regex)

    def extract(self, match: re.Match) -> dict[str, VarType]:
        result = {'.str': match.group()}
        for child in self.children:
            if child.group_name is not None and match.group(child.group_name):
                result.update(child.parse(match.group(child.group_name)))
        return result

    def parse(self, input_string: str) -> dict[str, int | list[str] | str]:
        return super().parse(textwrap.dedent(input_string))


def parse_template(template: str, input_string: str) -> dict[str, VarType]:
    return TemplateMatchItem(template).parse(input_string)


class MissingVariableError(Exception):
    def __init__(self, template: str, variables: dict[str, Any], missing_var: str):
        message = (
            'Missing variable in template serialization.\n\n'
            "Template:\n```\n{}\n```\n\nVariables:\n```\n{}\n```\n\nMissing Variable: '{}'\n".format(
                template, json.dumps(variables, indent=4), missing_var
            )
        )
        super().__init__(message)


def serialize_template(template: str, variables: dict[str, VarType]) -> str:
    def format_var(match: re.Match) -> str:
        var_name: str = match.group(1).split(':')[0].strip()
        if var_name + '.str' in variables:
            return variables[var_name + '.str']
        if var_name not in variables:
            raise MissingVariableError(template, variables, var_name)
        if not isinstance(variables[var_name], int | str):
            logger.warning('Formatting for list type not implemented ("{}" variable)', var_name)
        return str(variables[var_name])

    return re.sub(r'{{(.*?)}}', format_var, template)


def extract_data_vars(data: VarType) -> VarType:
    if isinstance(data, dict):
        filtered_data = {}
        for key, value in data.items():
            if not key.endswith('.str'):
                filtered_data[key] = extract_data_vars(value)
        return filtered_data
    if isinstance(data, list):
        return [extract_data_vars(item) for item in data]
    return data
